#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import locale
import os.path
from enum import Enum
from typing import Any, Dict, List, Optional

import oss2
from alibabacloud_credentials.client import Client as CredentialClient
from alibabacloud_credentials.exceptions import CredentialException
from alibabacloud_credentials.models import Config as CredentialConfig
from alibabacloud_credentials.providers import (
    CredentialsUriProvider,
    EcsRamRoleCredentialProvider,
    EnvironmentVariableCredentialsProvider,
    OIDCRoleArnCredentialProvider,
    ProfileCredentialsProvider,
    RamRoleArnCredentialProvider,
    RsaKeyPairCredentialProvider,
)
from alibabacloud_credentials.utils import auth_constant
from oss2.models import SimplifiedBucketInfo
from prompt_toolkit import prompt
from prompt_toolkit.validation import Validator

from ..common.logging import get_logger
from ..common.oss_utils import OssUriObj
from ..common.utils import (
    is_domain_connectable,
    make_list_resource_iterator,
    random_str,
)
from ..session import Session
from .helper.consts import REGION_INFOS
from .helper.utils import (
    UserProfile,
    WorkspaceRoles,
    confirm,
    localized_text,
    mask_and_trim,
    mask_secret,
    not_empty,
    print_highlight,
    print_warning,
    radio_list_prompt,
    validate_bucket_name,
)

logger = get_logger(__name__)

DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".pai", "config.json")
DEFAULT_CREDENTIAL_INI_PATH = os.path.join(
    os.path.expanduser("~"), ".alibabacloud", "credentials.ini"
)

REGION_ID_ENV_KEYS = [
    "ALIBABACLOUD_REGION_ID",
    "ALICLOUD_REGION_ID",
    "REGION",
    "dsw_region",
    "DSW_REGION",
]

local, _ = locale.getdefaultlocale()


CREDENTIAL_INI_TEMPLATE = """[default]
enable = true
type = access_key
access_key_id = {access_key_id}
access_key_secret = {access_key_secret}
"""


def _get_default_credential_client() -> Optional[CredentialClient]:
    try:
        return CredentialClient()
    except CredentialException:
        logger.debug("Not found credential from default credential provider chain.")


class CredentialProviderType(Enum):
    EnvironmentVariable = EnvironmentVariableCredentialsProvider
    OIDCRoleArn = OIDCRoleArnCredentialProvider
    EcsRamRole = EcsRamRoleCredentialProvider
    RamRoleArn = RamRoleArnCredentialProvider
    RsaKeyPair = RsaKeyPairCredentialProvider
    Profile = ProfileCredentialsProvider
    CredentialUri = CredentialsUriProvider

    @classmethod
    def get_current_provider(cls) -> Optional["CredentialProviderType"]:
        from alibabacloud_credentials.providers import DefaultCredentialsProvider

        d = {t.value: t for t in cls}
        provider = DefaultCredentialsProvider()
        for p in provider.user_configuration_providers:
            if p.get_credentials():
                return d.get(p.__class__)

    @classmethod
    def credential_hint(cls, cred_type: Optional["CredentialProviderType"]) -> str:
        provider_hints = {
            CredentialProviderType.EnvironmentVariable: localized_text(
                "The credential source is: Environment Variable",
                "凭证来源: 环境变量(ALIBABACLOUD_ACCESS_KEY_ID, ALIBABACLOUD_ACCESS_KEY_SECRET)",
            ),
            CredentialProviderType.OIDCRoleArn: localized_text(
                "The credential source is: OIDC Role Arn",
                "凭证来源: OIDC Role Arn",
            ),
            CredentialProviderType.EcsRamRole: localized_text(
                "The credential source is: ECS Ram Role",
                "凭证来源: ECS Ram Role",
            ),
            CredentialProviderType.RamRoleArn: localized_text(
                "The credential source is: Ram Role Arn",
                "凭证来源: Ram Role Arn",
            ),
            CredentialProviderType.RsaKeyPair: localized_text(
                "The credential source is: RSA Key Pair",
                "凭证来源: RSA Key Pair",
            ),
            CredentialProviderType.Profile: localized_text(
                "The credential source is: Profile",
                "凭证来源: Profile(~/.alibabacloud/credentials.ini)",
            ),
            CredentialProviderType.CredentialUri: localized_text(
                "The credential source is: CredentialUri (EnvironmentVairbale ALIBABA_CLOUD_CREDENTIALS_URI)",
                "凭证来源: CredentialUri (环境变量 ALIBABA_CLOUD_CREDENTIALS_URI)",
            ),
        }

        return provider_hints.get(
            cred_type,
            localized_text(
                "The credential source is: Unknown",
                "凭证来源: 未知",
            ),
        )


def prompt_for_credential():
    default_credential_client = _get_default_credential_client()
    if not default_credential_client:
        # Prompt for Access Key ID and Access Key Secret
        access_key_id = prompt(
            localized_text(
                "Please enter your Alibaba Cloud account AccessKeyId: ",
                "请输入您的阿里云账号AccessKeyId: ",
            ),
            validator=Validator.from_callable(
                validate_func=not_empty,
                error_message=localized_text(
                    "AccessKeyId can not be empty string.",
                    "AccessKeyId 不能为空",
                ),
                move_cursor_to_end=True,
            ),
        ).strip()
        access_key_secret = prompt(
            localized_text(
                "Please enter your Alibaba Cloud account AccessKeySecret: ",
                "请输入您的阿里云账号AccessKeySecret: ",
            ),
            is_password=True,
            validator=Validator.from_callable(
                validate_func=not_empty,
                error_message=localized_text(
                    "AccessKeySecret can not be empty string.",
                    "AccessKeySecret 不能为空",
                ),
                move_cursor_to_end=True,
            ),
        ).strip()
        credential_config = CredentialConfig(
            access_key_id=access_key_id,
            access_key_secret=access_key_secret,
            type=auth_constant.ACCESS_KEY,
        )
        credential_client = CredentialClient(config=credential_config)
    else:
        # Credential chain documentation:
        # https://help.aliyun.com/zh/sdk/developer-reference/v2-manage-python-access-credentials
        print(
            localized_text(
                "Use credential from default credential provider chain:",
                "使用默认的凭证链获取访问密钥:",
            )
        )

        credential_source_hint = CredentialProviderType.credential_hint(
            CredentialProviderType.get_current_provider()
        )
        print(credential_source_hint)
        credential_client = default_credential_client
        credential_config = None

    region_id = prompt_for_region()
    print_highlight(
        localized_text(
            "The current configuration of Credential and RegionId:",
            "当前配置的访问密钥和地域:",
        )
    )

    access_key_id = credential_client.get_access_key_id()
    access_key_secret = credential_client.get_access_key_secret()
    security_token = credential_client.get_security_token()

    print_highlight(f"AccessKeyId: {access_key_id}")
    print_highlight(f"AccessKeySecret: { mask_secret(access_key_secret)}")
    if security_token:
        print_highlight(f"SecurityToken: {mask_and_trim(security_token)}")
    print_highlight(f"RegionId: {region_id}")
    user_profile = UserProfile(
        credential_config=credential_config,
        region_id=region_id,
    )
    print_highlight(f"IdentityType: {user_profile.identify_type}")

    # Write input credential to default credential config file.
    if credential_config:
        raw_config = CREDENTIAL_INI_TEMPLATE.format(
            access_key_id=access_key_id,
            access_key_secret=access_key_secret,
        )
        os.makedirs(os.path.dirname(DEFAULT_CREDENTIAL_INI_PATH), exist_ok=True)
        with open(DEFAULT_CREDENTIAL_INI_PATH, "w") as f:
            f.write(raw_config)
        print(
            localized_text("Credential saved to: ", "密钥已保存至: ")
            + DEFAULT_CREDENTIAL_INI_PATH
        )

    check_product_authorization(user_profile)
    return user_profile


def prompt_for_region():
    for key in REGION_ID_ENV_KEYS:
        region_id = os.environ.get(key)
        if region_id:
            print(
                localized_text(
                    f"Config RegionId from environment variable({key}): {region_id} ",
                    f"从环境变量({key})中获取RegionId: {region_id}",
                )
            )
            return region_id

    region_name_map = {r["regionId"]: r["regionName"] for r in REGION_INFOS}
    region_row_format = "{:<30}{}"
    supported_region_ids = list(region_name_map.keys())

    region_list = [
        (
            region_id,
            region_row_format.format(
                region_id, localized_text(*region_name_map[region_id])
            ),
        )
        for region_id in supported_region_ids
    ]

    region_id = radio_list_prompt(
        title=localized_text(
            "Please select the region where the service will be used: ",
            "请选择使用服务的地域: ",
        ),
        values=region_list,
        erase_when_done=True,
    )
    return region_id


def check_product_authorization(user_profile: UserProfile):
    print(localized_text("Check Dependent Service Authorization：", "云产品依赖检查: "))
    auths = {
        auth["RamRoleName"]: auth["IsAuthorized"]
        for auth in user_profile.get_production_authorizations()
    }
    for role_name, authorized in auths.items():
        print(
            "{:<40}{}".format(
                role_name, "Authorized" if authorized else "NotAuthorized"
            )
        )


def prompt_for_workspace(user_profile: UserProfile):
    """Ask for workspace configurations."""
    workspace_api = user_profile.get_workspace_api()

    row_format = "{:<50}{}"

    def workspace_choice_name(workspace: Dict[str, Any]):
        if workspace.get("IsDefault"):
            return row_format.format(
                workspace["WorkspaceName"],
                str(workspace["WorkspaceId"]) + "  (*default)",
            )
        return row_format.format(
            workspace["WorkspaceName"],
            str(workspace["WorkspaceId"]),
        )

    items = sorted(
        [
            (item, workspace_choice_name(workspace=item))
            for item in make_list_resource_iterator(
                workspace_api.list,
                page_number=1,
                page_size=50,
            )
        ],
        key=lambda x: (x[0].get("IsDefault", False), x[0]["WorkspaceName"]),
        reverse=True,
    )
    if not items:
        if user_profile.is_account:
            raise RuntimeError(
                localized_text(
                    f"You do not have any available Workspace in the selected region ("
                    f"{user_profile.region_id}). Please create one in the region."
                    f" through the PAI console (https://pai.console.aliyun.com/).",
                    f"您在选择的Region ({user_profile.region_id}）下没有可用的工作空间，"
                    f"请通过PAI的控制台（ https://pai.console.aliyun.com/ ）创建一个工作空间。",
                )
            )
        else:
            raise RuntimeError(
                localized_text(
                    f"You do not have any available Workspace in the selected region ("
                    f"{user_profile.region_id}). Please create a new workspace or "
                    f"contact an administrator to be added to an existing workspace.",
                    f"您在选择的Region ({user_profile.region_id}) 下没有可用的工作空间，"
                    f"请创建一个新的工作空间，或是联系管理员将您添加到已有的工作空间。",
                )
            )
    else:
        header = row_format.format("WorkspaceName", "WorkspaceId")
        workspace_item = radio_list_prompt(
            localized_text(
                f"Please select the workspace you need to access:\n    {header}",
                f"请选择您需要访问的工作空间\n    {header})",
            ),
            values=items,
            erase_when_done=True,
        )

        workspace_id = workspace_item["WorkspaceId"]

    roles = user_profile.get_roles_in_workspace(workspace_id)
    role_info = ", ".join(roles)
    if not any(r for r in roles if r in WorkspaceRoles.recommend_roles()):
        print_warning(
            localized_text(
                f"We recommend using a workspace with at least developer permission. "
                f"Role in the workspace: {role_info}",
                f"建议使用至少有开发者权限的工作空间，当前工作空间角色：{role_info}",
            )
        )

    print_highlight(
        localized_text(
            "Current workspace configuration selected:",
            "当前选择的工作空间信息:",
        )
    )
    print_highlight(
        "WorkspaceName: {}\nWorkspaceId: {}\nRoles: {}".format(
            workspace_item["WorkspaceName"],
            workspace_item["WorkspaceId"],
            role_info,
        )
    )

    return workspace_id


def prompt_for_oss_bucket(user_profile: UserProfile, workspace_id: str):
    default_storage_uri, endpoint = user_profile.get_default_oss_storage_uri(
        workspace_id=workspace_id
    )
    print(
        localized_text(
            "Please configure the OSS Bucket to be used.", "请配置使用的OSS Bucket"
        )
    )
    if user_profile.is_ram_user:
        print_warning(
            localized_text(
                "Please confirm that current RAM user has been granted permission to"
                " list buckets and access the OSS Bucket that is needed.",
                "请确认当前的RAM账号 ListBuckets 权限以及 需要使用的OSS Bucket的读写权限。",
            )
        )

    bucket_name = None
    # use the default OSS storage URI configured in the Workspace.
    if default_storage_uri and confirm(
        localized_text(
            f"Do you want to use the default OSS Bucket of the current workspace"
            f" ({default_storage_uri})?",
            f"是否使用当前工作空间的默认OSS Bucket ({default_storage_uri}) ?",
        )
    ):
        bucket_name = OssUriObj(default_storage_uri).bucket_name

    if not bucket_name:
        buckets: List[SimplifiedBucketInfo] = user_profile.list_oss_buckets()
        if not buckets:
            print(
                localized_text(
                    f"You do not have any available OSS Bucket in the region"
                    f" ({user_profile.region_id}).",
                    f"您在当前Region （{user_profile.region_id}） 下没有可用的OSS Bucket。",
                )
            )
            bucket_name = prompt_for_create_oss_bucket(user_profile, workspace_id)
        else:
            index = radio_list_prompt(
                localized_text(
                    "Please select the OSS Bucket you want to use:",
                    "请选择您需要使用的OSS Bucket：",
                ),
                values=[(idx, b.name) for idx, b in enumerate(buckets)],
                erase_when_done=True,
            )
            bucket_name = buckets[index].name

    try:
        bucket_info = user_profile.get_bucket_info(bucket_name=bucket_name)
    except oss2.exceptions.AccessDenied:
        # try to get bucket info with ListBuckets API if the user has no permission to
        # GetBucketInfo API.
        buckets = user_profile.list_oss_buckets(prefix=bucket_name)
        bucket_info = next((b for b in buckets if b.name == bucket_name), None)

    if not bucket_info:
        print_warning(
            localized_text(
                "Failed to get bucket info, use default endpoint.",
                "获取 Bucket 信息失败，使用默认 Endpoint。",
            )
        )
        region_id = user_profile.region_id
        extranet_endpoint, intranet_endpoint = (
            f"oss-{region_id}.aliyuncs.com",
            f"oss-{region_id}-internal.aliyuncs.com",
        )
    else:
        extranet_endpoint, intranet_endpoint = (
            bucket_info.extranet_endpoint,
            bucket_info.intranet_endpoint,
        )

    # If Workspace has no default OSS storage URI and user has permission to edit,
    # prompt to set the default OSS storage URI.
    if not default_storage_uri and user_profile.has_permission_edit_config(
        workspace_id=workspace_id
    ):
        prompt_for_set_default_oss_storage(
            user_profile, workspace_id, bucket_name, intranet_endpoint=intranet_endpoint
        )

    row_format = "{:<60}{}"
    intra_endpoint_connectable = is_domain_connectable(intranet_endpoint, timeout=1)
    candidates = [
        (
            intranet_endpoint,
            row_format.format(
                intranet_endpoint,
                localized_text(
                    "Internal endpoint (Please use in PAI-DSW Notebook, ECS and other "
                    "intranet environment)",
                    "内网Endpoint(请在PAI-DSW Notebook, ECS等内网环境中使用)",
                ),
            ),
        ),
        (
            extranet_endpoint,
            row_format.format(
                extranet_endpoint,
                localized_text(
                    "Public endpoint",
                    "外网Endpoint",
                ),
            ),
        ),
    ]

    if not intra_endpoint_connectable:
        candidates = candidates[::-1]

    endpoint = radio_list_prompt(
        localized_text(
            "Please select the Endpoint to access OSS:",
            "请选择访问OSS使用的Endpoint：",
        ),
        values=candidates,
        erase_when_done=True,
    )
    return bucket_name, endpoint


def prompt_for_set_default_oss_storage(
    user_profile: UserProfile,
    workspace_id: str,
    bucket_name: str,
    intranet_endpoint: str,
):
    yes_no = confirm(
        localized_text(
            "Whether config the bucket as the default storage for the Workspace?",
            "是否将当前的OSS Bucket作为工作空间的默认存储？",
        )
    )
    if yes_no:
        user_profile.set_default_oss_storage(
            workspace_id, bucket_name, intranet_endpoint=intranet_endpoint
        )


def prompt_for_create_oss_bucket(user_profile: UserProfile, workspace_id):
    gen_bucket_name = f"pai-{user_profile.account_id}-{workspace_id}"

    while True:
        bucket_name = prompt(
            localized_text(
                "Please confirm the OSS Bucket name：\n",
                "请确认新建的 OSS 的 Bucket名称：\n",
            ),
            default=gen_bucket_name,
            validator=Validator.from_callable(
                validate_func=validate_bucket_name,
                error_message=localized_text(
                    "Invalid bucket name.",
                    "错误的Bucket名称.",
                ),
                move_cursor_to_end=True,
            ),
        )

        res = confirm(
            localized_text(
                f"Confirm to create the OSS Bucket. BucketName:{bucket_name}",
                f"是否使用该名称创建新的OSS Bucket。 BucketName:{bucket_name}",
            )
        )
        if not res:
            continue

        try:
            user_profile.create_oss_bucket(bucket_name)
            print_highlight(
                localized_text(
                    f"Create bucket succeeded. {bucket_name}",
                    f"Bucket创建成功. {bucket_name}",
                )
            )
            break
        except oss2.exceptions.ServerError as e:
            if e.code == "BucketAlreadyExists":
                # if bucket already exists.
                print_warning(
                    localized_text(
                        f"Bucket already exists: {bucket_name}",
                        f"当前Bucket已经被占用: {bucket_name}",
                    )
                )
                if bucket_name == gen_bucket_name:
                    gen_bucket_name = (
                        f"pai-{user_profile.account_id}-{workspace_id}-{random_str(6)}"
                    )
                continue
            raise e

    return bucket_name


def prompt_for_config_writing(
    user_profile: UserProfile,
    workspace_id,
    bucket_name,
    bucket_endpoint,
):
    # Print configuration summary.
    print_highlight(localized_text("Current configuration options:", "当前配置项:"))
    print_highlight(f"WorkspaceId: { workspace_id}")
    print_highlight(f"RegionId: {user_profile.region_id}")
    print_highlight(f"OSS Bucket Name: {bucket_name}")
    print_highlight(f"OSS Bucket Endpoint: {bucket_endpoint}")

    if os.path.exists(DEFAULT_CONFIG_PATH):
        overwrite = confirm(
            localized_text(
                f"The configuration file already exists. Do you want to overwrite the"
                f" original configuration? ({DEFAULT_CONFIG_PATH})",
                f"配置文件已经存在，是否覆盖原先配置 ({DEFAULT_CONFIG_PATH})？",
            )
        )
        if not overwrite:
            print_highlight(
                localized_text("Discard current configuration.", "放弃保存当前配置。")
            )
            return

    sess = Session(
        region_id=user_profile.region_id,
        workspace_id=workspace_id,
        oss_bucket_name=bucket_name,
        oss_endpoint=bucket_endpoint,
    )
    sess.save_config()
    print(
        localized_text(
            f"Configuration saved successfully: {DEFAULT_CONFIG_PATH}",
            f"配置保存成功: {DEFAULT_CONFIG_PATH}",
        )
    )


def prompt_config_with_default_dsw_role(user_profile: UserProfile):
    print(
        localized_text(
            "The current DSW instance is bound to the default PAI DSW role,"
            " automatically utilizes the instance's workspace and OSS Bucket configurations.",
            "当前DSW实例绑定PAI DSW默认角色，自动使用实例的工作空间和OSS Bucket配置",
        )
    )
    instance_id = os.environ.get("DSW_INSTANCE_ID")
    if not instance_id:
        raise RuntimeError(
            "Not found PAI DSW instance id from environment variable: DSW_INSTANCE_ID"
        )
    instance_info = user_profile.get_instance_info(instance_id=instance_id)
    workspace_id = instance_info["WorkspaceId"]
    workspace_name = instance_info["WorkspaceName"]
    user_id = instance_info["UserId"]
    roles = user_profile.get_roles_in_workspace(workspace_id, user_id)
    role_info = ", ".join(roles)
    print_highlight(
        localized_text(
            "Current workspace configuration:",
            "当前实例的工作空间信息:",
        )
    )
    print_highlight(
        "WorkspaceName: {}\nWorkspaceId: {}\nRoles: {}".format(
            workspace_name,
            workspace_id,
            role_info,
        )
    )

    default_storage_uri, endpoint = user_profile.get_default_oss_storage_uri(
        workspace_id=workspace_id,
    )

    if not default_storage_uri:
        print_warning(
            localized_text(
                "WARNING: The STS credential generated by the default ROLE only support accessing "
                "the default OSS Bucket storage of the workspace.\n"
                "It is not configured for the current workspace, please "
                "reference the document to configure the default OSS Bucket storage: \n"
                "https://help.aliyun.com/zh/pai/user-guide/manage-workspaces#section-afd-ntr-nwh",
                '警告：默认角色产生的STS凭证仅支持访问"工作空间默认存储"的OSS Bucket。\n'
                "当前工作空间没有配置默认OSS Bucket存储，请参考帮助文档进行配置：\n"
                "https://help.aliyun.com/zh/pai/user-guide/manage-workspaces#section-afd-ntr-nwh",
            )
        )
        bucket_name, endpoint = None, None
    else:
        bucket_name = OssUriObj(default_storage_uri).bucket_name
    return workspace_id, bucket_name, endpoint


def run():
    """
    The flow of pai config.

    1. Prompt for the basic user profile, including the AccessKey and region.
    2. Prompt for the workspace.
        2.1. If no workspace exists, raise a RuntimeError.
        2.2. List workspaces for the user to select.
    3. Prompt for the OSS Bucket.
        3.1. If the workspace has a default OSS storage, prompt whether to use the
            default one. If yes, use the default OSS Bucket, and go to 3.3.
        3.2. List Buckets for the user to select. If there are no OSS Buckets, prompt
            to create an OSS Bucket; otherwise, return the Bucket selected by the user.
        3.3. Prompt for the OSS endpoint for the Bucket.
    4. Prompt for configuration writing.

    """
    # Ask for Account profile
    user_profile = prompt_for_credential()

    if user_profile.is_dsw_default_role():
        (
            workspace_id,
            bucket_name,
            bucket_endpoint,
        ) = prompt_config_with_default_dsw_role(user_profile=user_profile)
    else:
        # Ask for workspace
        workspace_id = prompt_for_workspace(user_profile=user_profile)
        # Ask for OSS Bucket
        bucket_name, bucket_endpoint = prompt_for_oss_bucket(
            user_profile=user_profile, workspace_id=workspace_id
        )

    # Ask for config writing
    prompt_for_config_writing(
        user_profile,
        workspace_id,
        bucket_name,
        bucket_endpoint,
    )


if __name__ == "__main__":
    run()



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict

from ..common.logging import get_logger
from ..libs.alibabacloud_paiflow20210202.models import (
    CreatePipelineRequest,
    CreatePipelineResponseBody,
    GetPipelineResponseBody,
    GetPipelineSchemaResponseBody,
    ListPipelinesRequest,
    ListPipelinesResponseBody,
    UpdatePipelineRequest,
)
from .base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI

logger = get_logger(__name__)


class PipelineAPI(WorkspaceScopedResourceAPI):

    BACKEND_SERVICE_NAME = ServiceName.PAIFLOW

    _get_method = "get_pipeline_with_options"
    _get_schema_method = "get_pipeline_schema_with_options"
    _list_method = "list_pipelines_with_options"
    _create_method = "create_pipeline_with_options"
    _delete_method = "delete_pipeline_with_options"
    _update_method = "update_pipeline_with_options"
    _get_caller_provider_method = "get_caller_provider_with_options"

    def get(self, pipeline_id) -> Dict[str, Any]:
        resp: GetPipelineResponseBody = self._do_request(
            method_=self._get_method, pipeline_id=pipeline_id
        )
        return resp.to_map()

    def get_by_identifier(self, identifier, provider, version="v1"):
        res = self.list(identifier=identifier, provider=provider, version=version)
        if not res.items:
            return
        if len(res.items) > 1:
            logger.warning(
                f"Get pipeline by identifier returns more one pipeline: "
                f"identifier={identifier} provider={provider} version={version}"
            )
        return res.items[0]

    def get_schema(self, pipeline_id):
        resp: GetPipelineSchemaResponseBody = self._do_request(
            method_=self._get_schema_method, pipeline_id=pipeline_id
        )

        return resp.to_map()

    def list(
        self,
        identifier=None,
        provider=None,
        version=None,
        page_number=None,
        page_size=None,
    ) -> PaginatedResult:
        if provider:
            workspace_id = self.workspace_id_none_placeholder
        else:
            workspace_id = None

        request = ListPipelinesRequest(
            page_number=page_number,
            page_size=page_size,
            pipeline_provider=provider,
            pipeline_version=version,
            pipeline_identifier=identifier,
            workspace_id=workspace_id,
        )

        resp: ListPipelinesResponseBody = self._do_request(
            method_=self._list_method, request=request
        )
        return self.make_paginated_result(resp)

    def create(self, manifest):
        request = CreatePipelineRequest(manifest=manifest)
        resp: CreatePipelineResponseBody = self._do_request(
            method_=self._create_method, request=request
        )
        return resp.pipeline_id

    def delete(self, pipeline_id):
        self._do_request(
            method_=self._delete_method,
            pipeline_id=pipeline_id,
        )

    def update(self, pipeline_id, manifest):
        request = UpdatePipelineRequest(
            manifest=manifest,
        )
        self._do_request(
            method_=self._update_method, pipeline_id=pipeline_id, request=request
        )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class PAIException(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return "{}: {}".format(type(self).__name__, self.message)

    def __str__(self):
        return self.__repr__()


class PredictionException(PAIException):
    def __init__(self, code, message):
        super(PredictionException, self).__init__(message)
        self.code = code
        self.message = message

    def __repr__(self):
        return "{}: code={}, {}".format(type(self).__name__, self.code, self.message)


class UnexpectedStatusException(PAIException):
    """Raised when resource status is not expected."""

    def __init__(self, message, status):
        self.status = status
        super(UnexpectedStatusException, self).__init__(message)

    def __repr__(self):
        return "{}: status={}, {}".format(
            type(self).__name__, self.status, self.message
        )


class DuplicatedMountException(PAIException):
    """Raised if a OSS path is mounted twice."""


class MountPathIsOccupiedException(PAIException):
    """Raised if target mount path is already used."""



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import base64
import functools
import json
import posixpath
import time
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from io import IOBase
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

import aiohttp
import requests

from .common.consts import FrameworkTypes
from .common.docker_utils import ContainerRun
from .common.logging import get_logger
from .common.utils import http_user_agent, is_package_available
from .exception import PredictionException
from .serializers import (
    JsonSerializer,
    PyTorchSerializer,
    SerializerBase,
    TensorFlowSerializer,
)
from .session import Session, get_default_session

if is_package_available("openai"):
    from openai import OpenAI


logger = get_logger(__name__)

_PAI_SERVICE_CONSOLE_URI_PATTERN = (
    "https://pai.console.aliyun.com/?regionId={region_id}&workspaceId={workspace_id}#"
    "/eas/serviceDetail/{service_name}/detail"
)

_QUEUE_SERVICE_REQUEST_ID_HEADER = "X-Eas-Queueservice-Request-Id"
_QUEUE_SERVICE_SINK_PATH = "sink"
_DEFAULT_ASYNC_WORKER_COUNT = 30


class ServiceStatus(object):
    """All EAS inference service status."""

    Running = "Running"
    Waiting = "Waiting"
    Scaling = "Scaling"
    Stopped = "Stopped"
    Failed = "Failed"
    DeleteFailed = "DeleteFailed"

    @classmethod
    def completed_status(cls):
        return [
            cls.Running,
            cls.Stopped,
            cls.Failed,
            cls.DeleteFailed,
        ]


class EndpointType(object):
    # Public Internet Endpoint
    INTERNET = "INTERNET"

    # VPC Endpoint
    INTRANET = "INTRANET"


class ServiceType(object):
    Standard = "Standard"
    Async = "Async"


class PredictorBase(ABC):
    @abstractmethod
    def predict(self, *args, **kwargs) -> Any:
        """Perform inference on the provided data and return prediction result."""

    @abstractmethod
    def raw_predict(
        self,
        data: Any = None,
        path: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        method: str = "POST",
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs,
    ):
        pass


class RawResponse(object):
    """Response object returned by the predictor.raw_predict."""

    def __init__(self, status_code: int, headers: Dict[str, str], content: bytes):
        """Initialize a RawResponse object.

        Args:
            status_code (int):
            headers (dict):
            content (bytes):
        """
        self.status_code = status_code
        self.headers = headers
        self.content = content

    def json(self):
        """Returns the json-encoded content of a response

        Returns:
            Dict[str, Any]: The json-encoded content of a response.

        """
        return json.loads(self.content)


class _ServicePredictorMixin(object):
    def __init__(
        self,
        service_name: str,
        session: Optional[Session] = None,
        endpoint_type: str = EndpointType.INTERNET,
        serializer: Optional[SerializerBase] = None,
    ):
        self.service_name = service_name
        self.session = session or get_default_session()
        self._service_api_object = self.describe_service()
        self.endpoint_type = endpoint_type
        self.serializer = serializer or self._get_default_serializer()
        self._request_session = requests.Session()

    def __repr__(self):
        return "{}(service_name={}, endpoint_type={})".format(
            type(self).__name__,
            self.service_name,
            self.endpoint_type,
        )

    def __del__(self):
        self._request_session.close()

    def refresh(self):
        self._service_api_object = self.describe_service()

    @property
    def endpoint(self):
        if self.endpoint_type == EndpointType.INTRANET:
            return self._service_api_object["IntranetEndpoint"]
        else:
            return self._service_api_object["InternetEndpoint"]

    @property
    def intranet_endpoint(self):
        return self._service_api_object["IntranetEndpoint"]

    @property
    def internet_endpoint(self):
        return self._service_api_object["InternetEndpoint"]

    @property
    def service_status(self):
        """Returns the status of the service."""
        return self._service_api_object["Status"]

    @property
    def access_token(self) -> str:
        """Access token of the service."""
        return self._service_api_object["AccessToken"]

    @property
    def labels(self) -> Dict[str, str]:
        """Labels of the service."""
        labels = {
            item["LabelKey"]: item["LabelValue"]
            for item in self._service_api_object.get("Labels", [])
        }
        return labels

    @property
    def console_uri(self):
        """Returns the console URI of the service."""
        return _PAI_SERVICE_CONSOLE_URI_PATTERN.format(
            workspace_id=self.session.workspace_id,
            region_id=self.session.region_id,
            service_name=self.service_name,
        )

    def _get_default_serializer(self):
        """Get default serializer for the predictor by inspecting the service config."""
        from pai.model._model import _BuiltinProcessor

        service_config = json.loads(self._service_api_object["ServiceConfig"])
        processor_code = service_config.get("processor")

        # If the prediction service is serving with custom processor or custom
        # container, use JsonSerializer as default serializer.
        if not processor_code:
            return JsonSerializer()

        if processor_code in (
            _BuiltinProcessor.PMML,
            _BuiltinProcessor.XGBoost,
        ):
            return JsonSerializer()
        elif processor_code.startswith(FrameworkTypes.TensorFlow.lower()):
            serializer = TensorFlowSerializer()
            return serializer
        elif processor_code.startswith(FrameworkTypes.PyTorch.lower()):
            return PyTorchSerializer()
        else:
            return JsonSerializer()

    def _post_init_serializer(self):
        """Post-initialize the serializer by invoking serializer.inspect_from_service"""
        if not hasattr(self.serializer, "__post_init_serializer_flag") and hasattr(
            self.serializer, "inspect_from_service"
        ):
            self.serializer.inspect_from_service(
                self.service_name, session=self.session
            )
            setattr(self.serializer, "__post_init_serializer_flag", 1)

    def inspect_model_signature_def(self):
        """Get SignatureDef of the serving model.

        .. note::

            Only the service using the TensorFlow processor supports getting the
            model signature_definition.

        Returns:
            Dict[str, Any]: A dictionary representing the signature definition of the
                serving model.

        """
        service_config = json.loads(self._service_api_object["ServiceConfig"])
        processor_code = service_config.get("processor")

        if processor_code and processor_code.startswith("tensorflow"):
            return TensorFlowSerializer.inspect_model_signature_def(
                self.service_name, session=self.session
            )
        raise RuntimeError(
            "Only the online prediction service using the TensorFlow processor supports"
            " getting the signature_definition"
        )

    def describe_service(self) -> Dict[str, Any]:
        """Describe the service that referred by the predictor.

        Returns:
            Dict[str, Any]: Response from PAI API service.

        """
        return self.session.service_api.get(self.service_name)

    def start_service(self, wait=True):
        """Start the stopped service."""
        self.session.service_api.start(name=self.service_name)
        if wait:
            status = ServiceStatus.Running
            unexpected_status = ServiceStatus.completed_status()
            unexpected_status.remove(status)
            type(self)._wait_for_status(
                service_name=self.service_name,
                status=status,
                unexpected_status=unexpected_status,
                session=self.session,
            )
        self.refresh()

    def stop_service(self, wait=True):
        """Stop the running service."""
        self.session.service_api.stop(name=self.service_name)
        if wait:
            status = ServiceStatus.Stopped
            unexpected_status = ServiceStatus.completed_status()
            unexpected_status.remove(status)
            unexpected_status.remove(ServiceStatus.Running)

            type(self)._wait_for_status(
                service_name=self.service_name,
                status=status,
                unexpected_status=unexpected_status,
                session=self.session,
            )
        self.refresh()

    def delete_service(self):
        """Delete the service."""
        self.session.service_api.delete(name=self.service_name)

    def wait_for_ready(self):
        """Wait until the service enter running status.

        Raises:
            RuntimeError: Raise if the service terminated unexpectedly.

        """
        if self.service_status == ServiceStatus.Running:
            return

        logger.info(
            "Service waiting for ready: service_name={}".format(self.service_name)
        )
        unexpected_status = ServiceStatus.completed_status()
        unexpected_status.remove(ServiceStatus.Running)
        type(self)._wait_for_status(
            service_name=self.service_name,
            status=ServiceStatus.Running,
            unexpected_status=unexpected_status,
            session=self.session,
        )

        # hack: PAI-EAS gateway may not be ready when the service is ready.
        self._wait_for_gateway_ready()
        self.refresh()

    def wait(self):
        """Wait for the service to be ready."""
        return self.wait_for_ready()

    def _wait_for_gateway_ready(self, attempts: int = 60, interval: int = 2):
        """Hacky way to wait for the service gateway to be ready.

        Args:
            attempts (int): Number of attempts to wait for the service gateway to be
                ready.
            interval (int): Interval between each attempt.
        """

        def _is_gateway_ready():
            # can't use HEAD method to check gateway status because the service will
            # block the request until timeout.
            resp = self._send_request(method="GET")
            logger.debug(
                "Check gateway status result: status_code=%s content=%s",
                resp.status_code,
                resp.content,
            )
            res = not (
                # following status code and content indicates the gateway is not ready
                (
                    resp.status_code == 503
                    and (b"no healthy upstream" in resp.content or not resp.content)
                )
                or (resp.status_code == 404 and not resp.content)
            )
            return res

        err_count_threshold = 3
        err_count = 0
        while attempts > 0:
            attempts -= 1
            try:
                if _is_gateway_ready():
                    break
            except requests.exceptions.RequestException as e:
                err_count += 1
                if err_count >= err_count_threshold:
                    logger.warning("Failed to check gateway status: %s", e)
                    break
            time.sleep(interval)
        else:
            logger.warning("Timeout waiting for gateway to be ready.")

    @classmethod
    def _wait_for_status(
        cls,
        service_name: str,
        status: str,
        unexpected_status: List[str],
        interval: int = 3,
        session: Optional[Session] = None,
    ):
        session = session or get_default_session()
        service_api_object = session.service_api.get(service_name)
        last_status = service_api_object["Status"]
        last_msg = service_api_object["Message"]
        time.sleep(interval)

        while True:
            service_api_object = session.service_api.get(service_name)
            # Check the service status
            cur_status = service_api_object["Status"]
            if cur_status == status:
                return status
            elif unexpected_status and cur_status in unexpected_status:
                # Unexpected terminated status
                raise RuntimeError(
                    f"The Service terminated unexpectedly: "
                    f"name={service_api_object['ServiceName']} "
                    f"status={service_api_object['Status']} "
                    f"reason={service_api_object['Reason']} "
                    f"message={service_api_object['Message']}."
                )
            elif (
                last_status == cur_status and service_api_object["Message"] == last_msg
            ) and cur_status != ServiceStatus.Waiting:
                # If service.status and service.message have not changed and
                # service.status is not 'Waiting', do not print the service
                # status/message.
                pass
            else:
                logger.info(
                    f"Refresh Service status: "
                    f"name={service_api_object['ServiceName']} "
                    f"id={service_api_object['ServiceId']} "
                    f"status={service_api_object['Status']} "
                    f"reason={service_api_object['Reason']} "
                    f"message={service_api_object['Message']}."
                )

            last_status = service_api_object["Status"]
            last_msg = service_api_object["Message"]
            time.sleep(interval)

    def switch_version(self, version: int):
        """Switch service to target version.

        Args:
            version (int): Target version

        """
        service_api_object = self.describe_service()

        current_version = service_api_object["CurrentVersion"]
        latest_version = service_api_object["LatestVersion"]
        if current_version == version:
            raise ValueError("Target version equals to current version.")

        if version > latest_version:
            raise ValueError("Target version greater than latest version.")
        self.session.service_api.update_version(self.service_name, version=version)

    @classmethod
    def deploy(
        cls,
        config: Dict[str, Any],
        session: Optional[Session] = None,
        endpoint_type: str = EndpointType.INTERNET,
        serializer: Optional[SerializerBase] = None,
        wait: bool = True,
    ) -> PredictorBase:
        """Deploy an online prediction service using given configuration.

        Args:
            config (Dict[str, Any]): A dictionary of service configuration.
            session (:class:`pai.session.Session`, optional): An optional
                session object. If not provided, a default session will be used.
            serializer: An optional serializer object. If not provided, a
                default serializer will be used.
            endpoint_type: The type of endpoint to use.
            wait: Whether to wait for the service to be ready before returning.

        Returns:
            :class:`pai.predictor.PredictorBase`: A Predictor object for the deployed
                online prediction service.

        """
        session = session or get_default_session()
        name = session.service_api.create(config=config)

        if wait:
            # Wait until the service is ready
            unexpected_status = ServiceStatus.completed_status()
            unexpected_status.remove(ServiceStatus.Running)
            Predictor._wait_for_status(
                service_name=name,
                status=ServiceStatus.Running,
                unexpected_status=unexpected_status,
                session=session,
            )

        service_api_obj = session.service_api.get(name)

        if service_api_obj["ServiceType"] == ServiceType.Async:
            p = AsyncPredictor(
                service_name=name,
                endpoint_type=endpoint_type,
                serializer=serializer,
            )
        else:
            p = Predictor(
                service_name=name,
                endpoint_type=endpoint_type,
                serializer=serializer,
            )

        return p

    def _build_url(
        self, path: Optional[str] = None, params: Dict[str, str] = None
    ) -> str:
        url = self.endpoint
        if path:
            if path.startswith("/"):
                path = path[1:]
            url = posixpath.join(url, path)

        # Add params to URL
        url = url + "?" + urlencode(params) if params else url
        return url

    def _build_headers(self, headers: Dict[str, str] = None) -> Dict[str, str]:
        headers = headers or dict()
        headers["Authorization"] = self.access_token
        headers["User-Agent"] = http_user_agent(headers.get("User-Agent"))
        return headers

    def _handle_input(self, data):
        return self.serializer.serialize(data) if self.serializer else data

    def _handle_output(self, content: bytes):
        return self.serializer.deserialize(content) if self.serializer else content

    def _handle_raw_input(self, data):
        if isinstance(data, (IOBase, bytes, str)):
            # if data is a file-like object, bytes, or string, it will be sent as
            # request body
            json_data, data = None, data
        else:
            # otherwise, it will be treated as a JSON serializable object and sent as
            # JSON.
            json_data, data = data, None

        return json_data, data

    def _handle_raw_output(self, status_code: int, headers: dict, content: bytes):
        return RawResponse(status_code, headers, content)

    def _send_request(
        self,
        data=None,
        path=None,
        method="POST",
        json=None,
        headers=None,
        params=None,
        **kwargs,
    ):
        url = self._build_url(path)
        resp = self._request_session.request(
            url=url,
            json=json,
            data=data,
            headers=self._build_headers(headers),
            method=method,
            params=params,
            **kwargs,
        )
        return resp

    async def _send_request_async(
        self,
        data=None,
        path=None,
        method="POST",
        json=None,
        headers=None,
        params=None,
        **kwargs,
    ):
        url = self._build_url(path=path, params=params)
        headers = self._build_headers(headers)
        async with aiohttp.ClientSession() as session:
            return await session.request(
                method=method,
                url=url,
                headers=headers,
                data=data,
                json=json,
                **kwargs,
            )


class Predictor(PredictorBase, _ServicePredictorMixin):
    """Predictor is responsible for making prediction to an online service.

    The `predictor.predict` method sends the input data to the online prediction service
    and returns the prediction result. The serializer object of the predictor is
    responsible for data transformation when the `predict` method is invoked. The input
    data is serialized using the `serializer.serialize` method before it is sent, and
    the response is deserialized using the `serializer.deserialize` method before the
    prediction result returns.

    Examples::

        # Initialize a predictor object from an existing service using PyTorch
        # processor.
        torch_predictor = Predictor(service_name="example_torch_service")
        result = torch_predictor.predict(numpy.asarray([[22,33,44], [19,22,33]]))
        assert isinstance(result, numpy.ndarray)

    """

    def __init__(
        self,
        service_name: str,
        endpoint_type: str = EndpointType.INTERNET,
        serializer: Optional[SerializerBase] = None,
        session: Optional[Session] = None,
    ):
        """Construct a `Predictor` object using an existing prediction service.

        Args:
            service_name (str): Name of the existing prediction service.
            endpoint_type (str): Selects the endpoint used by the predictor, which
                should be one of `INTERNET` or `INTRANET`. The `INTERNET` endpoint type
                means that the predictor calls the service over a public endpoint, while
                the `INTRANET` endpoint type is over a VPC endpoint.
            serializer (SerializerBase, optional): A serializer object that transforms
                the input Python object for data transmission and deserialize the
                response data to Python object.
            session (Session, optional): A PAI session object used for communicating
                with PAI service.
        """
        super(Predictor, self).__init__(
            service_name=service_name,
            session=session or get_default_session(),
            endpoint_type=endpoint_type,
            serializer=serializer,
        )
        self._check()

    def _check(self):
        config = json.loads(self._service_api_object["ServiceConfig"])
        if config.get("metadata", {}).get("type") == ServiceType.Async:
            logger.warning(
                "Predictor is not recommended to make prediction to a async"
                " prediction service."
            )

    def predict(self, data):
        """Make a prediction with the online prediction service.

        The serializer object for the predictor is responsible for data transformation
        when the 'predict' method is invoked. The input data is serialized using the
        `serializer.serialize` method before it is sent, and the response is
        deserialized using the `serializer.deserialize` method before the prediction
        result returns.

        Args:
            data: The input data for the prediction. It will be serialized using the
                serializer of the predictor before transmitted to the prediction
                service.

        Returns:
            object: Prediction result.

        Raises:
            PredictionException: Raise if status code of the prediction response does
                not equal 2xx.
        """
        self._post_init_serializer()
        data = self._handle_input(data)
        resp = self._send_request(
            data,
        )
        if resp.status_code // 100 != 2:
            raise PredictionException(resp.status_code, resp.content)
        return self._handle_output(
            resp.content,
        )

    def raw_predict(
        self,
        data: Any = None,
        path: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        method: str = "POST",
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs,
    ) -> RawResponse:
        """Make a prediction with the online prediction service.

        Args:
            data (Any): Input data to be sent to the prediction service. If it is a
                file-like object, bytes, or string, it will be sent as the request body.
                Otherwise, it will be treated as a JSON serializable object and sent as
                JSON.
            path (str, optional): Path for the request to be sent to. If it is provided,
                it will be appended to the endpoint URL (Default None).
            headers (dict, optional): Request headers.
            method (str, optional): Request method, default to 'POST'.
            timeout(float, tuple(float, float), optional): Timeout setting for the
                request (Default 10).
            **kwargs: Additional keyword arguments for the request.
        Returns:
            RawResponse: Prediction response from the service.

        Raises:
            PredictionException: Raise if status code of the prediction response does
                not equal 2xx.
        """
        json_data, data = self._handle_raw_input(data)
        resp = self._send_request(
            data=data,
            json=json_data,
            method=method,
            path=path,
            headers=headers,
            timeout=timeout,
            **kwargs,
        )
        if resp.status_code // 100 != 2:
            raise PredictionException(resp.status_code, resp.content)

        resp = RawResponse(
            status_code=resp.status_code,
            content=resp.content,
            headers=dict(resp.headers),
        )
        return resp

    def openai(self, url_suffix: str = "v1", **kwargs) -> "OpenAI":
        """Initialize an OpenAI client from the predictor.

        Only used for OpenAI API compatible services, such as Large Language Model
        service from PAI QuickStart.

        Args:
            url_suffix (str, optional): URL suffix that will be appended to the
                EAS service endpoint to form the base URL for the OpenAI client.
                (Default "v1").
            **kwargs: Additional keyword arguments for the OpenAI client.

        Returns:
            OpenAI: An OpenAI client object.
        """

        if not is_package_available("openai"):
            raise ImportError(
                "openai package is not installed, install it with `pip install openai`."
            )

        if url_suffix.startswith("/"):
            default_base_url = posixpath.join(self.endpoint, url_suffix[1:])
        else:
            default_base_url = posixpath.join(self.endpoint, url_suffix)
        base_url = kwargs.pop("base_url", default_base_url)
        api_key = kwargs.pop("api_key", self.access_token)

        return OpenAI(base_url=base_url, api_key=api_key, **kwargs)


class WaitConfig(object):
    """WaitConfig is used to set polling configurations for waiting for asynchronous
    requests to complete."""

    def __init__(self, max_attempts: int = 0, interval: int = 5):
        if interval <= 0:
            raise ValueError("interval must be positive integer.")
        self.max_attempts = max_attempts
        self.interval = interval


class AsyncTask(object):
    """AsyncTask is a wrapper class for `concurrent.futures.Future` object that represents
    a prediction call submitted to an async prediction service.
    """

    def __init__(
        self,
        future: Future,
    ):
        self.future = future
        super(AsyncTask, self).__init__()

    def result(self, timeout: Optional[float] = None):
        """
        Returns the prediction result of the call.

        Args:
            timeout (float, optional): Timeout setting  (Default None).

        Returns:
            The result of the prediction call.

        """
        return self.future.result(timeout=timeout)

    def done(self):
        return self.future.done()

    def exception(self, timeout: Optional[float] = None) -> Optional[Exception]:
        return self.future.exception()

    def running(self):
        return self.future.running()

    def cancel(self):
        return self.future.cancel()

    def cancelled(self):
        return self.future.cancelled()


class AsyncPredictor(PredictorBase, _ServicePredictorMixin):
    """A class that facilitates making predictions to asynchronous prediction service.

    Examples::

        # Initialize an AsyncPredictor object using the name of a running service.
        async_predictor = AsyncPredictor(service_name="example_service")

        # Make a prediction with the service and get the prediction result.
        resp = async_predictor.predict(data="YourPredictionData")
        result = resp.wait()

        # Make a prediction with async API.
        import asyncio
        result = asyncio.run(async_predictor.predict_async(data="YourPredictionData"))

    """

    def __init__(
        self,
        service_name: str,
        max_workers: Optional[int] = None,
        endpoint_type: str = EndpointType.INTERNET,
        serializer: Optional[SerializerBase] = None,
        session: Optional[Session] = None,
    ):
        """Construct a `AsyncPredictor` object using an existing async prediction service.

        Args:
            service_name (str): Name of the existing prediction service.
            max_workers (int): The maximum number of threads that can be used to
                execute the given prediction calls.
            endpoint_type (str): Selects the endpoint used by the predictor, which
                should be one of `INTERNET` or `INTRANET`. The `INTERNET` endpoint type
                means that the predictor calls the service over a public endpoint, while
                the `INTRANET` endpoint type is over a VPC endpoint.
            serializer (SerializerBase, optional): A serializer object that transforms
                the input Python object for data transmission and deserialize the
                response data to Python object.
            session (Session, optional): A PAI session object used for communicating
                with PAI service.
        """

        super(AsyncPredictor, self).__init__(
            service_name=service_name,
            session=session or get_default_session(),
            endpoint_type=endpoint_type,
            serializer=serializer,
        )
        self._max_workers = max_workers
        self.executor = ThreadPoolExecutor(max_workers=self._max_workers)
        self._check()

    @property
    def max_workers(self):
        return self._max_workers

    @max_workers.setter
    def max_workers(self, n: int):
        if hasattr(self, "executor"):
            logger.info("Waiting for all submitted tasks in the queue to complete...")
            self.executor.shutdown()
        self._max_workers = n
        self.executor = ThreadPoolExecutor(max_workers=self._max_workers)

    def __del__(self):
        """wait for all pending tasks to complete before exit."""
        if hasattr(self, "executor"):
            logger.info("Waiting for all pending tasks to complete...")
            self.executor.shutdown()
        super(AsyncPredictor, self).__del__()

    def _check(self):
        config = json.loads(self._service_api_object["ServiceConfig"])
        if config.get("metadata", {}).get("type") != ServiceType.Async:
            logger.warning(
                "AsyncPredictor is not recommended to make prediction to a standard "
                " prediction service."
            )

    def _get_result(
        self, request_id: str
    ) -> Optional[Tuple[int, Dict[str, str], bytes]]:
        resp = self._send_request(
            method="GET",
            path=_QUEUE_SERVICE_SINK_PATH,
            params={
                "requestId": request_id,
                # _raw_ is false because we want to get the encapsulated prediction
                # result in response body.
                "_raw_": "false",
            },
        )
        logger.debug(
            "Poll prediction result: request_id=%s status_code=%s, content=%s",
            request_id,
            resp.status_code,
            resp.content,
        )
        if resp.status_code == 204:
            # Status code 204 means could not find prediction response for the specific
            # request id.
            return

        # Raise exception if status code is not 2xx.
        if resp.status_code // 100 != 2:
            raise RuntimeError(
                "Pulling prediction result failed: status_code={} content={}".format(
                    resp.status_code, resp.content.decode("utf-8")
                )
            )
        return self._parse_encapsulated_response(resp.json()[0])

    def _parse_encapsulated_response(self, data) -> Tuple[int, Dict[str, str], bytes]:
        tags = data["tags"]
        # If the status code from prediction service is not 200, a tag with
        # key 'lastCode' will be added to the tags in response.
        status_code = int(tags.get("lastCode", 200))
        data = base64.b64decode(data["data"])
        # currently, headers are not supported in async prediction service.
        headers = dict()
        return status_code, headers, data

    async def _get_result_async(
        self, request_id: str
    ) -> Optional[Tuple[int, Dict[str, str], bytes]]:
        resp = await self._send_request_async(
            method="GET",
            path=_QUEUE_SERVICE_SINK_PATH,
            params={
                "requestId": request_id,
                # _raw_ is false because we want to get the encapsulated prediction
                # result in response body.
                "_raw_": "false",
            },
        )
        status_code = resp.status
        content = await resp.read()
        logger.debug(
            "Get prediction result: request_id=%s status_code=%s, content=%s",
            request_id,
            status_code,
            content,
        )
        if status_code == 204:
            # Status code 204 means could not find prediction response for the specific
            # request id.
            return
        if status_code // 100 != 2:
            raise RuntimeError(
                "Pulling prediction result failed: status_code={} content={}".format(
                    status_code, content.decode("utf-8")
                )
            )
        data = (await resp.json())[0]
        return self._parse_encapsulated_response(data)

    def _poll_result(
        self, request_id: str, wait_config: WaitConfig
    ) -> Tuple[int, Dict[str, str], bytes]:
        # if max_attempts is negative or zero, then wait forever
        attempts = -1 if wait_config.max_attempts <= 0 else wait_config.max_attempts
        while attempts != 0:
            attempts -= 1
            result = self._get_result(request_id=request_id)
            if not result:
                time.sleep(wait_config.interval)
                continue
            status_code, headers, content = result
            # check real prediction response
            if status_code // 100 != 2:
                raise PredictionException(
                    code=status_code,
                    message=f"Prediction failed: status_code={status_code}"
                    f" content={content.decode()}",
                )
            return status_code, headers, content

        # Polling prediction result timeout.
        raise RuntimeError(
            f"Polling prediction result timeout: request_id={request_id}, "
            f"total_time={wait_config.max_attempts * wait_config.interval}"
        )

    async def _poll_result_async(
        self, request_id, wait_config: WaitConfig
    ) -> Tuple[int, Dict[str, str], bytes]:
        # if max_attempts is negative or zero, then wait forever
        attempts = -1 if wait_config.max_attempts <= 0 else wait_config.max_attempts
        while attempts != 0:
            attempts -= 1
            result = await self._get_result_async(request_id)
            if not result:
                await asyncio.sleep(wait_config.interval)
                continue
            status_code, headers, content = result
            # check real prediction response
            if status_code // 100 != 2:
                raise PredictionException(
                    f"Prediction failed: status_code={status_code} content={content.decode()}"
                )
            return status_code, headers, content

        # Polling prediction result timeout.
        raise RuntimeError(
            f"Polling prediction result timeout: request_id={request_id}, "
            f"total_time={wait_config.max_attempts * wait_config.interval}"
        )

    def _get_request_id(self, resp: requests.models.Response) -> str:
        if resp.status_code // 100 != 2:
            raise RuntimeError(
                f"Send prediction request failed. status_code={resp.status_code} "
                f"message={resp.text}"
            )

        if _QUEUE_SERVICE_REQUEST_ID_HEADER not in resp.headers:
            logger.error(
                f"Send prediction request failed. Missing request id."
                f" status_code={resp.status_code} content={resp.text}"
            )
            raise RuntimeError("Missing request id in response header.")

        request_id = resp.headers[_QUEUE_SERVICE_REQUEST_ID_HEADER]
        logger.debug(
            f"Send prediction request successfully. request_id={request_id}"
            f" status_code={resp.status_code}",
        )
        return request_id

    async def _get_request_id_async(self, resp: aiohttp.ClientResponse) -> str:
        content = await resp.read()
        if resp.status != 200:
            raise RuntimeError(
                "Send request to async prediction service failed: status_code={} "
                "content={}".format(resp.status, content.decode("utf-8"))
            )

        if _QUEUE_SERVICE_REQUEST_ID_HEADER not in resp.headers:
            logger.error(
                f"Send prediction request failed. Missing request id."
                f" status_code={resp.status} content={content.decode()}"
            )
            raise RuntimeError("Missing request id in response header.")
        request_id = resp.headers[_QUEUE_SERVICE_REQUEST_ID_HEADER]
        logger.debug(
            f"Send prediction request successfully. request_id={request_id}"
            f" status_code={resp.status}",
        )
        return request_id

    def _predict_fn(
        self,
        data,
    ):
        """Make a prediction with the async prediction service."""
        # serialize input data
        data = self._handle_input(data)
        resp = self._send_request(data=data)
        request_id = self._get_request_id(resp)
        logger.debug("Async prediction RequestId: ", request_id)
        # poll prediction result
        status, headers, content = self._poll_result(
            request_id=request_id, wait_config=WaitConfig()
        )

        return self._handle_output(content)

    def _wrap_callback_fn(self, cb: Callable):
        """Wrap the callback function to handle the prediction result."""

        @functools.wraps(cb)
        def _(future: Future):
            return cb(future.result())

        return _

    def predict(
        self,
        data,
        callback: Optional[Union[Callable, List[Callable]]] = None,
    ):
        """Make a prediction with the async prediction service.

        The input data is serialized using the `serializer.serialize` method before it
        is sent, and the response body is deserialized using the
        `serializer.deserialize` method the prediction result returns.

        Args:
            data: The input data for the prediction. It will be serialized using the
                serializer of the predictor before transmitted to the prediction
                service.
            callback (Union[Callable, List[Callable]], optional): A Callback function,
                or a list of callback functions used to process the prediction result.

        Returns:
            AsyncTask: The task object that can be used to retrieve the prediction
                result.
        """
        self._post_init_serializer()
        future = self.executor.submit(self._predict_fn, data)

        if isinstance(callback, Callable):
            callback = [callback]

        if callback:
            for cb in callback:
                future.add_done_callback(self._wrap_callback_fn(cb))

        return AsyncTask(future=future)

    async def predict_async(self, data, wait_config: WaitConfig = WaitConfig()):
        """Make a prediction with the async prediction service.

        The serializer object for the predictor is responsible for data transformation
        when the 'predict' method is invoked. The input data is serialized using the
        `serializer.serialize` method before it is sent, and the response is
        deserialized using the `serializer.deserialize` method before the prediction
        result returns.

        Args:
            data: The input data for the prediction. It will be serialized using the
                serializer of the predictor before transmitted to the prediction
                service.
            wait_config (WaitConfig): A config object that controls the behavior of
                polling the prediction result.

        Returns:
            Prediction result.

        """
        self._post_init_serializer()
        data = self._handle_input(data)
        resp = await self._send_request_async(data=data)
        request_id = await self._get_request_id_async(resp)

        status_code, headers, content = await self._poll_result_async(
            request_id=request_id, wait_config=wait_config
        )
        return self._handle_output(content)

    def _raw_predict_fn(self, data, method, path, headers, **kwargs):
        json_data, data = self._handle_raw_input(data)
        resp = self._send_request(
            path=path,
            json=json_data,
            data=data,
            headers=self._build_headers(headers),
            method=method,
            **kwargs,
        )
        request_id = self._get_request_id(resp)
        status, headers, content = self._poll_result(
            request_id, wait_config=WaitConfig()
        )
        return RawResponse(status, headers, content)

    def raw_predict(
        self,
        data: Any = None,
        callback: Optional[Union[Callable, List[Callable], None]] = None,
        method: str = "POST",
        path: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> AsyncTask:
        """Make a prediction with the online prediction service.

        Args:
            data (Any): Input data to be sent to the prediction service. If it is a
                file-like object, bytes, or string, it will be sent as the request body.
                Otherwise, it will be treated as a JSON serializable object and sent as
                JSON.
            callback (Union[Callable, List[Callable]], optional): A Callback function,
                or a list of callback functions used to process the prediction result.
            path (str, optional): Path for the request to be sent to. If it is provided,
                it will be appended to the endpoint URL (Default None).
            headers (dict, optional): Request headers.
            method (str, optional): Request method, default to 'POST'.
            **kwargs: Additional keyword arguments for the request.
        Returns:
            AsyncTask: The task object that can be used to retrieve the prediction
                result.

        Examples:

            from pai.predictor import AsyncPredictor, AsyncTask

            predictor = AsyncPredictor()
            task: AsyncTask = predictor.raw_predict(data="YourPredictionData")
            print(task.result())

        """

        future = self.executor.submit(
            self._raw_predict_fn, data, method, path, headers, **kwargs
        )
        cbs = [callback] if isinstance(callback, Callable) else callback
        if cbs:
            for cb in cbs:
                future.add_done_callback(self._wrap_callback_fn(cb))

        return AsyncTask(future=future)

    async def raw_predict_async(
        self,
        data,
        wait_config: WaitConfig = WaitConfig(),
        method: str = "POST",
        headers: Optional[Dict[str, str]] = None,
        path: Optional[str] = None,
        **kwargs,
    ) -> RawResponse:
        """Make a prediction with the online prediction service.

        Args:
            data (Any): Input data to be sent to the prediction service. If it is a
                file-like object, bytes, or string, it will be sent as the request body.
                Otherwise, it will be treated as a JSON serializable object and sent as
                JSON.
            wait_config (WaitConfig): A config object that controls the behavior of
                polling the prediction result.
            path (str, optional): Path for the request to be sent to. If it is provided,
                it will be appended to the endpoint URL (Default None).
            headers (dict, optional): Request headers.
            method (str, optional): Request method, default to 'POST'.
            **kwargs: Additional keyword arguments for the request.
        Returns:
            RawResponse: Prediction result.

        """
        if self.service_status not in ServiceStatus.completed_status():
            self.wait_for_ready()
        json_data, data = self._handle_raw_input(data)

        resp = await self._send_request_async(
            data=data,
            method=method,
            json=json_data,
            path=path,
            headers=headers,
            **kwargs,
        )
        request_id = await self._get_request_id_async(resp)
        # Polling the prediction result.
        status_code, headers, content = await self._poll_result_async(
            request_id=request_id, wait_config=wait_config
        )
        return self._handle_raw_output(status_code, headers, content)


class LocalPredictor(PredictorBase):
    """Perform prediction to a local service running with docker."""

    def __init__(
        self,
        port: int,
        container_id: Optional[str] = None,
        serializer: Optional[SerializerBase] = None,
    ):
        """LocalPredictor initializer.

        Args:
            port (int): The port of the local service.
            container_id (str, optional): The container id of the local service.
            serializer (SerializerBase, optional): A serializer object that transforms.
        """
        self.container_id = container_id
        self.port = port
        self.serializer = serializer or JsonSerializer()
        self._container_run = (
            self._build_container_run(container_id, port=port)
            if self.container_id
            else None
        )

    @classmethod
    def _build_container_run(cls, container_id, port):
        try:
            import docker
        except ImportError:
            raise ImportError("Please install docker first: pip install docker")
        client = docker.from_env()
        container = client.containers.get(container_id)

        return ContainerRun(container=container, port=port)

    def predict(self, data) -> Any:
        """Perform prediction with the given data.

        Args:
            data: The data to be predicted.
        """
        request_data = self.serializer.serialize(data=data)
        response = requests.post(
            url="http://127.0.0.1:{port}/".format(port=self._container_run.port),
            data=request_data,
        )

        if response.status_code // 100 != 2:
            raise PredictionException(
                code=response.status_code,
                message=response.content,
            )

        return self.serializer.deserialize(response.content)

    def _build_headers(
        self, headers: Optional[Dict[str, str]] = None
    ) -> Dict[str, str]:
        headers = headers or dict()
        headers["User-Agent"] = http_user_agent(headers.get("User-Agent"))
        return headers

    def _build_url(self, path: Optional[str] = None):
        url = "http://127.0.0.1:{}".format(self.port)
        if path:
            if path.startswith("/"):
                path = path[1:]
            url = posixpath.join(url, path)
        return url

    def raw_predict(
        self,
        data: Any = None,
        path: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        method: str = "POST",
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs,
    ) -> RawResponse:
        """Make a prediction with the online prediction service.

        Args:
            data (Any): Input data to be sent to the prediction service. If it is a
                file-like object, bytes, or string, it will be sent as the request body.
                Otherwise, it will be treated as a JSON serializable object and sent as
                JSON.
            path (str, optional): Path for the request to be sent to. If it is provided,
                it will be appended to the endpoint URL (Default None).
            headers (dict, optional): Request headers.
            method (str, optional): Request method, default to 'POST'.
            timeout(float, tuple(float, float), optional): Timeout setting for the
                request (Default 10).
        Returns:
            RawResponse: Prediction response from the service.

        Raises:
            PredictionException: Raise if status code of the prediction response does
                not equal 2xx.
        """
        if isinstance(data, (IOBase, bytes, str)):
            # if data is a file-like object, bytes, or string, it will be sent as
            # request body
            json_data, data = None, data
        else:
            # otherwise, it will be treated as a JSON serializable object and sent as
            # JSON.
            json_data, data = data, None
        header = self._build_headers(headers=headers)
        url = self._build_url(path)
        resp = requests.request(
            url=url,
            json=json_data,
            data=data,
            headers=header,
            method=method,
            timeout=timeout,
            **kwargs,
        )
        resp = RawResponse(
            status_code=resp.status_code,
            content=resp.content,
            headers=dict(resp.headers),
        )
        if resp.status_code // 100 != 2:
            raise PredictionException(resp.status_code, resp.content)
        return resp

    def delete_service(self):
        """Delete the docker container that running the service."""
        if self._container_run:
            self._container_run.stop()

    def wait_for_ready(self):
        self._container_run.wait_for_ready()
        # ensure the server is ready.
        self._wait_local_server_ready()
        time.sleep(5)

    def wait(self):
        return self.wait_for_ready()

    def _wait_local_server_ready(
        self,
        interval: int = 5,
    ):
        """Wait for the local model server to be ready."""
        container_run = self._container_run
        while True:
            try:
                # Check whether the container is still running.
                if not container_run.is_running():
                    raise RuntimeError(
                        "Container exited unexpectedly, status: {}".format(
                            container_run.status
                        )
                    )

                # Make a HEAD request to the server, just test for connection.
                requests.head(
                    f"http://127.0.0.1:{container_run.port}/",
                )
                break
            except requests.ConnectionError:
                # ConnectionError means server is not ready.
                logger.debug("Waiting for the container to be ready...")
                time.sleep(interval)
                continue



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
import uuid
from abc import ABCMeta, abstractmethod

import six

from ...common.logging import get_logger
from ...common.utils import random_str
from ...common.yaml_utils import dump as yaml_dump
from ...common.yaml_utils import dump_all as yaml_dump_all
from ...session import get_default_session
from ..types import IO_TYPE_INPUTS, IO_TYPE_OUTPUTS, InputsSpec, OutputsSpec

logger = get_logger(__name__)


DEFAULT_PIPELINE_API_VERSION = "core/v1"


class ComponentBase(six.with_metaclass(ABCMeta, object)):
    def __init__(
        self,
        inputs,
        outputs,
    ):
        self._inputs = (
            inputs if isinstance(inputs, InputsSpec) else InputsSpec(inputs or [])
        )
        for input in self._inputs:
            input.bind(self, IO_TYPE_INPUTS)
        self._outputs = (
            outputs if isinstance(outputs, OutputsSpec) else OutputsSpec(outputs or [])
        )
        for output in self._outputs:
            output.bind(self, IO_TYPE_OUTPUTS)

    @property
    def inputs(self):
        """Inputs Spec of the operator.

        Returns:
            pai.pipeline.types.spec.InputsSpec: Inputs of the operator.

        """
        return self._inputs

    @property
    def outputs(self):
        """Outputs Spec of the operator.

        Returns:
            pai.pipeline.types.spec.OutputsSpec: Outputs of the operator

        """
        return self._outputs

    def translate_arguments(self, args):
        parameters, artifacts = [], []
        if not args:
            return parameters, artifacts

        requires = set([af.name for af in self.inputs.artifacts if af.required])
        not_supply = requires - set(args.keys())
        if len(not_supply) > 0:
            raise ValueError(
                "Required arguments is not supplied:%s" % ",".join(not_supply)
            )

        name_var_mapping = {
            item.name: item for item in itertools.chain(self.inputs, self.outputs)
        }

        for name, arg in args.items():
            if name not in name_var_mapping:
                logger.error(
                    "Provider useless argument:%s, it is not require by the pipeline manifest spec."
                    % name
                )
                raise ValueError("provided argument is not required:%s" % name)

            variable = name_var_mapping[name]
            value = variable.translate_argument(arg)

            if variable.variable_category == "artifacts":
                artifacts.append(value)
            else:
                parameters.append(value)

        repeated_artifacts = [
            af
            for af in itertools.chain(self.inputs.artifacts, self.outputs.artifacts)
            if af.repeated and af.count
        ]
        for af in repeated_artifacts:
            if af.name not in args:
                artifacts.append({"name": af.name, "value": [None] * af.count})

        return parameters, artifacts

    def set_artifact_count(self, artifact_name, count):
        """Set the count of repeated artifact in operator run.

        Args:
            artifact_name: output repeated artifact name.
            count:
        """
        artifacts = {
            item.name: item
            for item in itertools.chain(self.outputs.artifacts, self.inputs.artifacts)
        }
        artifact = artifacts.get(artifact_name)
        if not artifact:
            raise ValueError("artifact is not exists: %s" % artifact_name)

        if not artifact.repeated:
            raise ValueError("artifact is not repeated: %s", artifact_name)

        artifact.count = count
        return self

    def spec_to_dict(self):
        spec = {"inputs": self.inputs.to_dict(), "outputs": self.outputs.to_dict()}
        return spec

    def to_dict(self):
        data = {
            "apiVersion": DEFAULT_PIPELINE_API_VERSION,
            "metadata": self.metadata_to_dict(),
            "spec": self.spec_to_dict(),
        }
        return data

    def run(
        self, job_name=None, wait=True, arguments=None, show_outputs=True, **kwargs
    ):
        """Run the operator using the definition in SavedOperator and given arguments.

        Args:
            job_name (str): Name of the submit pipeline run job.
            arguments (dict): Inputs arguments used in the run workflow.
            wait (bool): Wait util the job stop(succeed or failed or terminated).
            show_outputs (bool): Show the outputs of the job.

        Returns:
            pai.pipeline.run.PipelineRun: PipelineRun instance of the submit job.

        """
        from pai.pipeline import PipelineRun

        parameters, artifacts = self.translate_arguments(arguments)
        pipeline_args = {
            "parameters": parameters,
            "artifacts": artifacts,
        }

        run_id = self._submit(job_name=job_name, args=pipeline_args)

        run_instance = PipelineRun.get(run_id=run_id)
        if not wait:
            return run_instance
        run_instance.wait_for_completion(show_outputs=show_outputs)
        return run_instance

    def as_step(self, name=None, inputs=None, depends=None):
        """Create a PipelineStep instance using the operator."""
        from pai.pipeline import PipelineStep

        return PipelineStep(
            inputs=inputs,
            component=self,
            name=name,
            depends=depends,
        )

    def as_loop_step(self, name, items, parallelism=None, inputs=None, depends=None):
        """Create a LoopStep instance using the operator."""
        from pai.pipeline.step import LoopStep

        return LoopStep(
            component=self,
            name=name,
            parallelism=parallelism,
            items=items,
            inputs=inputs,
            depends=depends,
        )

    def as_condition_step(self, name, condition, inputs=None, depends=None):
        """Create a conditional step using the operator."""
        from pai.pipeline.step import ConditionStep

        return ConditionStep(
            component=self,
            name=name,
            condition=condition,
            inputs=inputs,
            depends=depends,
        )


class UnRegisteredComponent(six.with_metaclass(ABCMeta, ComponentBase)):
    def __init__(self, inputs, outputs):
        super(UnRegisteredComponent, self).__init__(inputs=inputs, outputs=outputs)
        self._guid = uuid.uuid4().hex
        self._name = "tmp-{}".format(random_str(16))

    @property
    def guid(self):
        return self._guid

    def metadata_to_dict(self):
        # Hack: PAIFlow Service require field name in metadata dict
        return {
            "name": self._name,
            "guid": self._guid,
        }

    @property
    def name(self):
        return self._name

    def save(self, identifier, version):
        """Save the Pipeline in PAI service for reuse or share it with others.

        By specific the identifier, version and upload the manifest, the PipelineTemplate instance
        is store into the remote service and return the pipeline_id of the saved PipelineTemplate.
        Account UID in Alibaba Cloud is use as the provider of the saved operator by default.
        Saved PipelineTemplate could be fetch using the pipeline_id or the specific
        identifier-provider-version.

        Args:
            identifier (str): The identifier of the saved pipeline.
            version (str): Version of the saved pipeline.

        Returns:
            pai.pipeline.SavedTemplate: Saved PipelineTemplate instance
            (with pipeline_id generate by remote service).

        """
        from pai.pipeline.component import RegisteredComponent

        if not identifier or not version:
            raise ValueError(
                "Please provide the identifier and version for the operator."
            )

        manifest = self.to_manifest(identifier=identifier, version=version)

        session = get_default_session()
        pipeline_id = session.pipeline_api.create(manifest)
        return RegisteredComponent.get(pipeline_id)

    @classmethod
    def _patch_metadata(cls, pipeline_spec):
        if isinstance(pipeline_spec, dict):
            pipeline_spec["metadata"]["identifier"] = "tmp-%s" % random_str(16)
            pipeline_spec["metadata"]["version"] = "v0"
        elif isinstance(pipeline_spec, list):
            pipeline_spec[-1]["metadata"]["identifier"] = "tmp-%s" % random_str(16)
            pipeline_spec[-1]["metadata"]["version"] = "v0"
        return pipeline_spec

    def _submit(self, job_name, args):
        from pai.pipeline.run import PipelineRun

        session = get_default_session()
        pipeline_spec = self._patch_metadata(self.to_dict())
        if isinstance(pipeline_spec, dict):
            manifest = yaml_dump(pipeline_spec)
        # A Pipeline spec may contain unregistered operators. Such a pipeline spec is a list format.
        elif isinstance(pipeline_spec, list):
            manifest = yaml_dump_all(pipeline_spec)
        else:
            raise ValueError(
                "No support pipeline spec value type: %s" % (type(pipeline_spec))
            )

        run_id = PipelineRun.run(
            job_name,
            args,
            no_confirm_required=True,
            manifest=manifest,
            session=session,
        )
        return run_id

    def io_spec_to_dict(self):
        return {
            "inputs": self.inputs.to_dict(),
            "outputs": self.outputs.to_dict(),
        }

    @abstractmethod
    def to_manifest(self, identifier, version):
        pass

    def export_manifest(self, file_path, identifier, version):
        manifest = self.to_manifest(identifier=identifier, version=version)
        with open(file_path, "w") as f:
            f.write(manifest)



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

from collections import Counter, defaultdict

from ..common.logging import get_logger
from ..common.yaml_utils import dump as yaml_dump
from ..common.yaml_utils import dump_all as yaml_dump_all
from ..session import get_default_session
from .component._base import UnRegisteredComponent
from .types import InputsSpec, OutputsSpec, PipelineParameter, PipelineVariable
from .types.artifact import PipelineArtifact, PipelineArtifactElement

logger = get_logger(__name__)


class Pipeline(UnRegisteredComponent):
    """Represents pipeline instance in PAI Machine Learning pipeliner service.

    Pipeline can be constructed from multiple pipeline steps, or single container implementation.
    It is shareable and reusable workflow, present as YAML format in backend pipeline service.

    """

    def __init__(self, steps, inputs=None, outputs=None, **kwargs):
        """Pipeline initializer."""
        from pai.pipeline import PipelineStep

        # check parameter steps.
        if not steps:
            raise ValueError("Required at least one step in the pipeline")
        if not isinstance(steps, list) or any(
            x for x in steps if not isinstance(x, PipelineStep)
        ):
            raise ValueError(
                "Parameter steps must be a list of PipelineStep instances."
            )

        steps, inputs, outputs, unregistered_ops = self._build_pipeline(
            steps, inputs, outputs
        )

        self._steps = steps
        self._unregistered_components = unregistered_ops
        super(Pipeline, self).__init__(inputs=inputs, outputs=outputs, **kwargs)

    @property
    def steps(self):
        return self._steps

    def _build_pipeline(self, steps, inputs, outputs):
        """

        Args:
            steps:
            outputs:
            inputs:

        Returns:

        """
        steps, inputs, _ = self._infer_pipeline_graph(steps, inputs, outputs)
        inputs_spec = InputsSpec(inputs) if isinstance(inputs, list) else inputs
        outputs_spec = OutputsSpec(self._build_outputs(outputs))

        self._check_inputs_outputs_name_conflict(
            inputs_spec=inputs_spec, outputs_spec=outputs_spec
        )
        unregistered_ops = self._get_unregistered_components(steps)

        self._update_steps(steps)

        return (
            steps,
            inputs_spec,
            outputs_spec,
            unregistered_ops,
        )

    @classmethod
    def _get_unregistered_components(cls, steps):
        """Get the unregistered components used by the step in the pipeline.

        Args:
            steps: Steps in the pipeline.

        Returns:
            List[OperatorBase]: Return unregistered components using in the pipeline.
        """

        unregistered_ops = [
            step.component for step in steps if not step.is_component_registered
        ]
        seen = set()
        return [
            op
            for op in unregistered_ops
            if op.guid not in seen and not seen.add(op.guid)
        ]

    @classmethod
    def _infer_pipeline_inputs(cls, input):
        pipeline_inputs = set()
        if isinstance(input, PipelineArtifact):
            sources = []
            if input.repeated and input.value:
                sources = [item for item in input.value]
            elif not input.repeated and input.from_:
                sources = [input.from_]

            for item in sources:
                if isinstance(item, PipelineArtifact) and not item.parent:
                    pipeline_inputs.add(item)
                elif (
                    isinstance(item, PipelineArtifactElement)
                    and not item.artifact.parent
                ):
                    pipeline_inputs.add(item.artifact)
        elif (
            input.from_
            and isinstance(input.from_, PipelineVariable)
            and not input.from_.parent
        ):
            pipeline_inputs.add(input.from_)
        return pipeline_inputs

    @classmethod
    def _set_step_artifact_count(cls, steps):
        """The expanded count of the repeated artifact should be fixed while using in the pipeline.

        Args:
            steps: All the steps used in the pipeline.
        """
        for step in steps:
            for input in step.inputs.artifacts:
                sources = []
                if input.repeated and input.value:
                    sources = input.value
                elif not input.repeated and input.from_:
                    sources = [input.from_]
                artifact_elements = [
                    i
                    for i in sources
                    if isinstance(i, PipelineArtifactElement) and i.artifact.parent
                ]

                for elem in artifact_elements:
                    if elem.artifact.count is None or elem.index >= elem.artifact.count:
                        elem.artifact.count = elem.index + 1

    @classmethod
    def _infer_pipeline_graph(cls, steps, inputs, outputs):
        """Inference the DAG graph of pipeline by pipelines inputs, steps and outputs. The
         function walks through the pipeline graph bottom-up from outputs and steps,
          finding out all the required steps and inputs of the pipeline.

        Args:
            steps: steps used in the pipeline graph.
            inputs: inputs used in the pipeline.
            outputs: outputs definition of the pipeline.

        Returns:
            Tuple: Returns all the required steps, inputs and outputs.

        """
        inputs = inputs or []

        outputs = outputs or []
        if isinstance(outputs, dict):
            outputs = list(outputs.values())

        # find out all steps in the pipeline by topological sort.
        steps = steps or []
        visited_steps = set(
            steps + [output.parent for output in outputs if output.parent]
        )
        cur_steps = visited_steps.copy()
        while cur_steps:
            next_steps = set()
            for step in cur_steps:
                for output in step.outputs.artifacts:
                    if output.repeated:
                        output.reset_count()
                for depend in step.depends:
                    if depend not in visited_steps:
                        next_steps.add(depend)
                        visited_steps.add(depend)
            cur_steps = next_steps

        # infer the pipeline inputs from step inputs.
        infer_inputs = set()
        for step in visited_steps:
            for ipt in step.inputs:
                infer_inputs |= cls._infer_pipeline_inputs(ipt)

        cls._set_step_artifact_count(visited_steps)

        if inputs:
            if len(infer_inputs) != len(inputs):
                raise ValueError(
                    "Please provide complete pipeline inputs list: expected=%s, given=%s"
                    % (len(infer_inputs), len(inputs))
                )

            unexpected = [ipt.name for ipt in inputs if ipt not in infer_inputs]
            if unexpected:
                raise ValueError(
                    "Do not provide inputs which is not used in pipeline: %s"
                    % unexpected
                )
        else:
            inputs = sorted(
                list(infer_inputs),
                key=lambda x: 0 if x.variable_category == "parameters" else 1,
            )
        sorted_steps = cls._topo_sort(visited_steps)
        cls._check_steps(steps)

        return sorted_steps, inputs, outputs

    @classmethod
    def _build_outputs(cls, outputs):
        outputs = outputs or []
        if isinstance(outputs, dict):
            items = outputs.items()
        elif isinstance(outputs, (list, OutputsSpec)):
            items = [(item.name, item) for item in outputs]
        else:
            raise ValueError("Require list or dict, unexpected type:%s" % type(outputs))

        results = []
        for name, item in items:
            if isinstance(item, PipelineArtifact):
                results.append(
                    PipelineArtifact(
                        repeated=item.repeated,
                        name=name,
                        from_=item,
                        metadata=item.metadata,
                    )
                )
            elif isinstance(item, PipelineArtifactElement):
                results.append(
                    PipelineArtifact(
                        name=name,
                        from_=item,
                        metadata=item.artifact.metadata,
                    )
                )
            elif isinstance(item, PipelineParameter):
                results.append(PipelineParameter(name=name, typ=item.typ, from_=item))
            else:
                raise ValueError("Unexpected output type: %s", type(item))
        return results

    @classmethod
    def _topo_sort(cls, steps):
        rev_depends = defaultdict(set)
        for step in steps:
            for depend_step in step.depends:
                rev_depends[depend_step].add(step)
        # entry steps
        visited_steps = [step for step in steps if not step.depends]

        cur_steps = visited_steps
        while cur_steps:
            next_steps = []
            for step in cur_steps:
                for candidate_step in rev_depends[step]:
                    if candidate_step in next_steps or candidate_step in visited_steps:
                        continue
                    if all(s in visited_steps for s in candidate_step.depends):
                        next_steps.append(candidate_step)
            visited_steps.extend(next_steps)
            cur_steps = next_steps

        if len(visited_steps) != len(steps):
            raise ValueError("Cycle dependency detected, please check the input steps")

        return visited_steps

    @classmethod
    def _check_steps(cls, steps):
        """
        1. check if cycle dependency exists in pipeline DAG.
        2. check if step name conflict.
        3. naming the unnamed step.

        Args:
            steps:
        """

        if any(step.parent for step in steps):
            raise ValueError("Pipeline step has been used")

        step_names = [step.name for step in steps if step.name]
        conflicts = [k for k, v in Counter(step_names).items() if v > 1]
        if conflicts:
            raise ValueError(
                "Given pipeline step name conflict:%s" % ",".join(conflicts)
            )

    @classmethod
    def _check_inputs_outputs_name_conflict(cls, inputs_spec, outputs_spec):
        names = [v.name for v in inputs_spec.items + outputs_spec.items]
        conflicts = [name for name, c in Counter(names).items() if c > 1]
        if conflicts:
            raise ValueError(
                "Given input/output variable name conflict:%s" % ",".join(conflicts)
            )

    def _update_steps(self, steps):
        """
        Set the name of step and bind the step to the pipeline.

        Args:
            steps: List of steps in pipeline.
        """
        used_names = set([s.name for s in steps])
        used_names = set(used_names)
        for step in steps:
            step.parent = self
            if not step.name:
                step.name = self._gen_step_name(
                    step, used_names=used_names, search_limit=len(steps)
                )
                used_names.add(step.name)

    @classmethod
    def _gen_step_name(cls, step, used_names, search_limit=100):
        for i in range(search_limit):
            candidate = "%s-%s" % (step.gen_name_prefix(), i)
            if candidate not in used_names:
                return candidate
        raise ValueError("No available name for the step")

    @property
    def ref_name(self):
        return ""

    def validate_step_name(self, name):
        if name in self.steps:
            raise ValueError("Pipeline step name conflict: %s" % name)
        return name

    def dot(self):
        try:
            from graphviz import Digraph
        except ImportError:
            raise ImportError(
                "Unable to display pipeline, install graphviz first: "
                "pip install graphviz"
            )

        graph = Digraph()
        for step in self.steps:
            graph.node(step.name)
            for head in step.depends:
                graph.edge(head.name, step.name)
        return graph

    def to_dict(self, identifier=None, version=None):
        entrypoint = super(Pipeline, self).to_dict()

        if identifier is not None:
            entrypoint["metadata"]["identifier"] = identifier
        if version is not None:
            entrypoint["metadata"]["version"] = version
        if get_default_session():
            entrypoint["metadata"]["provider"] = get_default_session().provider

        entrypoint["spec"]["pipelines"] = [step.to_dict() for step in self.steps]
        if not self._unregistered_components:
            return entrypoint

        res = [op.to_dict() for op in self._unregistered_components]
        res.append(entrypoint)

        return res

    def to_manifest(self, identifier, version):
        d = self.to_dict(identifier, version)
        if isinstance(d, list):
            return yaml_dump_all(d)
        return yaml_dump(d)



import os
import posixpath
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from .common.logging import get_logger
from .common.oss_utils import OssUriObj, is_oss_uri
from .common.utils import is_dataset_id, is_nas_uri
from .libs.alibabacloud_pai_dsw20220101.models import (
    GetInstanceRequest,
    GetInstanceResponse,
    GetInstanceResponseBody,
    UpdateInstanceRequest,
    UpdateInstanceRequestDatasets,
)
from .session import get_default_session

logger = get_logger()


class OptionType(str, Enum):
    """
    The type of options for mounting data sources in DSW (Data Science Workshop).

    This enum defines different mounting modes for OSS (Object Storage Service)
    and custom datasets in DSW. Each mode has specific characteristics and use cases:

    Attributes:
        FastReadWrite: Allows fast read and write operations. Suitable for training
            data and models, but may have data consistency issues with concurrent
            read/write operations. Not recommended for work directories.

        IncrementalReadWrite: Ensures data consistency for incremental writes,
            but may have consistency issues when overwriting existing data.
            Slightly slower read speed. Suitable for saving training model weights.

        ConsistentReadWrite: Maintains data consistency during concurrent read/write
            operations. Suitable for scenarios requiring high data consistency
            but can tolerate slower read speeds. Ideal for saving code projects.

        ReadOnly: Allows only read operations, no writing permitted.
            Suitable for mounting public datasets.

    The choice of option type affects the underlying Jindo configuration for
    mounting OSS data in DSW. Users can select the appropriate mode based on
    their specific use case and performance requirements.

    """

    FastReadWrite = "FastReadWrite"
    IncrementalReadWrite = "IncrementalReadWrite"
    ConsistentReadWrite = "ConsistentReadWrite"
    ReadOnly = "ReadOnly"


def _default_instance() -> "DswInstance":
    """
    Get the default DSW Instance.

    Returns:
        DswInstance: The default DSW Instance.
    """
    instance_id = os.getenv("DSW_INSTANCE_ID")
    if not instance_id:
        raise RuntimeError(
            "Environment variable 'DSW_INSTANCE_ID' is not set, please check if you are running in DSW environment"
        )
    return DswInstance(instance_id)


def mount(
    source: str,
    mount_point: str = None,
    options: Optional[Dict[str, Any]] = None,
    option_type: Optional[OptionType] = None,
) -> str:
    """
    Dynamic mount a dataset to the DSW Instance.

    Args:
        source (str): The source to be mounted, can be a dataset id or an OSS uri.
        mount_point (str): Target mount point in the instance, if not specified, the
            mount point be generate with given source under the default mount point.
        options (dict): Options that apply to when mount a data source, can not be
            specified with option_type.
        option_type(str): Preset data source mount options, can not be specified with
            options.

    Returns:
        str: The mount point of the data source.

    Examples:
        >>> mount_point = mount("oss://my-bucket/my-object/path/to/dir")
        >>> mount_point = mount("oss://my-bucket/my-object/path/to/dir", mount_point="/my-target", option_type=OptionType.FastReadWrite)
        >>> mount_point = mount("oss://my-bucket/my-object/path/to/dir", mount_point="/my-target", options={
            "fs.oss.download.thread.concurrency": "8", # CPU core count * 2
            "fs.oss.upload.thread.concurrency": "8", # CPU core count * 2
            # jindo args, refer to https://help.aliyun.com/zh/emr/emr-on-ecs/user-guide/user-guide-of-jindofuse
            "fs.jindo.args": "-oattr_timeout=3 -oentry_timeout=0 -onegative_timeout=0 -oauto_cache -ono_symlink" # jindo args
        })


    """
    instance = _default_instance()
    return instance.mount(
        source,
        mount_point,
        options=options,
        option_type=option_type,
    )


def unmount(mount_point: str) -> None:
    """
    Unmount a dynamic mount dataset from the DSW Instance.

    Dynamic mount dataset is a special dataset in DSW, which is used to mount data sources
    dynamically when the instance is running. When unmount it, the dataset will be removed
    from the instance.


    Args:
        mount_point (str): The mount point to be unmounted.

    Returns:
        None
    """
    instance = _default_instance()
    instance.unmount(mount_point)


def list_dataset_configs(dynamic_only: bool = False) -> List[Dict[str, Any]]:
    """
    List all the datasets available in the DSW Instance.

    Args:
        dynamic_only (bool): Whether to list only the dynamic mount datasets.

    Returns:
        list: A list of dataset details.
    """
    instance = _default_instance()
    datasets = instance._get_instance_info().datasets
    if dynamic_only:
        datasets = [ds for ds in datasets if ds.dynamic]
    return [ds.to_map() for ds in datasets]


def default_dynamic_mount_path():
    """Get the default dynamic mount path of the DSW Instance.

    Returns:
        str: The default dynamic mount path of the DSW Instance.
    """
    instance = _default_instance()
    return instance.default_dynamic_mount_path()


def get_dynamic_mount_config() -> Dict[str, Any]:
    """
    Get the dynamic mount config of the DSW Instance.

    Returns:
        dict: The dynamic mount config of the DSW Instance.
    """
    instance = _default_instance()
    return instance.get_dynamic_mount_config()


class DswInstance:
    """A object representing a DSW notebook instance"""

    def __init__(self, instance_id: str):
        self.instance_id = instance_id
        self._instance_info: GetInstanceResponseBody = self._get_instance_info()

    def _get_instance_info(self):
        session = get_default_session()
        resp: GetInstanceResponse = session._acs_dsw_client.get_instance(
            self.instance_id,
            request=GetInstanceRequest(),
        )
        return resp.body

    def get_dynamic_mount_config(self):
        """Get the dynamic mount config of the DSW Instance.

        Returns:
            dict: The dynamic mount config of the DSW Instance.
        """
        return self._instance_info.dynamic_mount.to_map()

    def default_dynamic_mount_path(self) -> Optional[str]:
        """Get the default dynamic mount path of the DSW Instance.

        Returns:
            str: The default dynamic mount path of the DSW Instance.
        """
        if not self._instance_info.dynamic_mount.enable:
            return None
        if not self._instance_info.dynamic_mount.mount_points:
            return "/mnt/dynamic"
        return self._instance_info.dynamic_mount.mount_points[0].root_path

    def mount(
        self,
        source: str,
        mount_point: str = None,
        options: Union[str] = None,
        option_type: Union[OptionType] = None,
    ) -> str:
        """
        Dynamic mount a data source to the DSW Instance.

        Args:
            source (str): The source to be mounted, can be a dataset id or OSS/NAS uri.
            mount_point (str): Target mount point in the instance, if not specified, the
                mount point be generated with given source under the default mount point.
            options (str): Options that apply to when mount a data source, can not be
                specified with option_type.
            option_type(str): Preset data source mount options, can not be specified with
                options.

        Returns:
            str: The mount point of the data source.
        """
        if options and option_type:
            raise ValueError(
                "options and option_type cannot be specified at the same time"
            )
        if not self._instance_info.dynamic_mount.enable:
            raise RuntimeError(
                "Dynamic mount is not enabled for the DSW instance: {}".format(
                    self.instance_id
                )
            )

        sess = get_default_session()
        default_root_path = self.default_dynamic_mount_path()

        if is_oss_uri(source):
            obj = OssUriObj(source)
            if not obj.endpoint:
                obj.endpoint = sess.oss_endpoint or sess._get_default_oss_endpoint()
            # ensure mount source OSS uri is a directory
            _, dir_path, _ = obj.parse_object_key()
            uri = f"oss://{obj.bucket_name}.{obj.endpoint}{dir_path}"
            dataset_id = None
            dataset_version = None
        elif is_nas_uri(source):
            uri = source
            dataset_id = None
            dataset_version = None
        else:
            dataset_id = source
            uri = None
            if "/" in dataset_id:
                dataset_id, dataset_version = dataset_id.split("/", 1)
            else:
                dataset_version = "v1"

        if (
            not is_oss_uri(source)
            and not is_nas_uri(source)
            and not is_dataset_id(source)
        ):
            raise ValueError("Source must be oss uri or nas uri or dataset id")

        if not mount_point:
            if is_oss_uri(source):
                obj = OssUriObj(source)
                mount_point = f"{obj.bucket_name}/{obj.object_key}"
            elif is_nas_uri(source):
                raise ValueError("Mount point is required for nas url.")
            else:
                mount_point = source
        if not posixpath.isabs(mount_point):
            mount_point = posixpath.join(default_root_path, mount_point)

        resp: GetInstanceResponse = sess._acs_dsw_client.get_instance(
            self.instance_id, request=GetInstanceRequest()
        )
        datasets = [
            UpdateInstanceRequestDatasets().from_map(ds.to_map())
            for ds in resp.body.datasets
        ]
        datasets.append(
            UpdateInstanceRequestDatasets(
                dataset_id=dataset_id,
                dataset_version=dataset_version,
                dynamic=True,
                mount_path=mount_point,
                option_type=option_type,
                options=options,
                uri=uri,
            )
        )
        request = UpdateInstanceRequest(
            datasets=datasets,
        )
        update_resp = sess._acs_dsw_client.update_instance(
            instance_id=self.instance_id, request=request
        )
        print("Mount succeed, request id: {}".format(update_resp.body.request_id))
        return mount_point

    def unmount(self, mount_point: str) -> None:
        """
        Unmount a data source from the DSW Instance.

        Args:
            mount_point (str): The mount point to be unmounted.

        Returns:
            None
        """
        sess = get_default_session()

        resp: GetInstanceResponse = sess._acs_dsw_client.get_instance(
            self.instance_id, request=GetInstanceRequest()
        )
        datasets = [
            UpdateInstanceRequestDatasets().from_map(ds.to_map())
            for ds in resp.body.datasets
        ]

        unmount_ds = [ds for ds in datasets if ds.mount_path == mount_point]

        if not unmount_ds:
            raise ValueError(f"Not found dataset to unmount: {mount_point}")
        if len(unmount_ds) > 1:
            raise RuntimeError(f"Found multiple datasets to unmount: {mount_point}")

        dataset = unmount_ds[0]
        if not dataset.dynamic:
            raise ValueError(f"Dataset is not a dynamic mount point: {mount_point}")

        request = UpdateInstanceRequest(
            datasets=[
                UpdateInstanceRequestDatasets().from_map(ds.to_map())
                for ds in resp.body.datasets
                if ds.mount_path != mount_point
            ]
        )
        if not request.datasets:
            request.disassociate_datasets = True
        update_resp = sess._acs_dsw_client.update_instance(
            instance_id=self.instance_id, request=request
        )
        print("Unmount succeed, request id: {}".format(update_resp.body.request_id))



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

REGION_INFOS = [
    {
        "regionId": "cn-beijing",
        "regionName": [
            "China (Beijing)",
            "华北 2 （北京）",
        ],
    },
    {
        "regionId": "cn-hangzhou",
        "regionName": [
            "China (Hangzhou)",
            "华东 1 （杭州）",
        ],
    },
    # {
    #     "regionId": "cn-zhangjiakou",
    #     "regionName": [
    #         "China (Zhangjiakou)",
    #         "华北 3 （张家口）",
    #     ],
    # },
    {
        "regionId": "cn-shanghai",
        "regionName": [
            "China (Shanghai)",
            "华东 2 （上海）",
        ],
    },
    {
        "regionId": "cn-wulanchabu",
        "regionName": [
            "China (Wulanchabu)",
            "华北6（乌兰察布）",
        ],
    },
    {
        "regionId": "cn-shenzhen",
        "regionName": [
            "China (Shenzhen)",
            "华南 1 （深圳）",
        ],
    },
    {
        "regionId": "cn-hongkong",
        "regionName": [
            "China (Hongkong)",
            "中国香港",
        ],
    },
    {
        "regionId": "ap-southeast-1",
        "regionName": [
            "Singapore",
            "新加坡",
        ],
    },
    # {
    #     "regionId": "ap-southeast-2",
    #     "regionName": [
    #         "Australia (Sydney)",
    #         "澳大利亚（悉尼）",
    #     ],
    # },
    {
        "regionId": "ap-southeast-3",
        "regionName": [
            "Malaysia (Kuala Lumpur)",
            "马来西亚（吉隆坡）",
        ],
    },
    {
        "regionId": "ap-southeast-5",
        "regionName": [
            "Indonesia (Jakarta)",
            "印度尼西亚（雅加达）",
        ],
    },
    # {
    #     "regionId": "ap-south-1",
    #     "regionName": [
    #         "India (Mumbai)",
    #         "印度（孟买）",
    #     ],
    # },
    # {
    #     "regionId": "us-east-1",
    #     "regionName": [
    #         "US (Virginia)",
    #         "美国东部 1 （维吉尼亚）",
    #     ],
    # },
    # {
    #     "regionId": "us-west-1",
    #     "regionName": [
    #         "US (Silicon Valley)",
    #         "美国西部 1 （硅谷）",
    #     ],
    # },
    {
        "regionId": "eu-central-1",
        "regionName": [
            "Germany (Frankfurt)",
            "德国（法兰克福）",
        ],
    },
    # {
    #     "regionId": "me-east-1",
    #     "regionName": [
    #         "UAE (Dubai)",
    #         "阿联酋（迪拜）",
    #     ],
    # },
]
